/*
 * J2ME Compass
 * Copyright (C) 2006 Dana Peters
 * http://www.qcontinuum.org/compass
 * 
 * Changed 2009 by Vedran Vidovic.
 * Removed henson.midp.Float class for performance reasons.
 * Removed code unused by SgTimer.
 */

package org.qcontinuum.astro;

import java.util.Date;

public class RiseSetTime {
    
    final static private double sinh0 = 
        Math.sin(Math.toRadians(-50.0 / 60.0)); // Sunrise at h=-50'

    private double mRiseTime;
    private double mSetTime;
    private boolean mRises;
    private boolean mSets;

    public RiseSetTime(Date gmtDate, int minutesOffset, EarthPosition earth) {
        this(new UtcDate(gmtDate).getMJD(), -minutesOffset / 60.0, earth);
    }
    
    public RiseSetTime(double MJD, double zone, EarthPosition earth) {
        double phi = Math.toRadians(earth.getLatitude());
        double lambda = Math.toRadians(earth.getLongitude());
        double Cphi = Math.cos(phi);
        double Sphi = Math.sin(phi);

        int hour = 1;
        double root1 = 0;
        double root2 = 0;
        double xe, ye;
        int nRoot;
        double MJD0h = Math.floor(MJD + zone / 24.0) - zone / 24.0;

        // Initialize for search
        double y_minus = SinAlt(MJD0h, hour - 1, lambda, Cphi, Sphi) - sinh0;

        mRises = false;
        mSets  = false;

        // loop over search intervals from [0h-2h] to [22h-24h]
        do {
            double y_0 = SinAlt(MJD0h, hour, lambda, Cphi, Sphi) - sinh0;
            double y_plus = SinAlt(MJD0h, hour + 1, lambda, Cphi, Sphi) - sinh0;

            // find parabola through three values y_minus,y_0,y_plus
            nRoot = 0;

            // Coefficients of interpolating parabola y=a*x^2+b*x+c
            double a  = 0.5 * (y_plus + y_minus) - y_0;
            double b  = 0.5 * (y_plus - y_minus);
            double c  = y_0;

            // Find extreme value
            xe = -b / (2.0 * a);
            ye = (a * xe + b) * xe + c;
            double dis = b * b - 4.0 * a * c; // Discriminant of y=a*x^2+b*x+c

            if (dis >= 0) // Parabola has roots 
            {
                // dx = 0.5 * sqrt (dis) / fabs (a);
                double dx = 0.5 * Math.sqrt(dis) / Math.abs(a);
                root1 = xe - dx; 
                root2 = xe + dx;

                if (Math.abs(root1) <= 1)
                    ++nRoot;  
                if (Math.abs(root2) <= 1)
                    ++nRoot;
                if (root1 < -1)
                    root1 = root2;
            }

            if (nRoot == 1)
            {
                if (y_minus < 0)
                {
                    mRiseTime = root1 + hour;
                    mRises = true;
                } else {
                    mSetTime  = root1 + hour;
                    mSets  = true;
                }
            }

            if (nRoot == 2)
            {
                if (ye < 0)
                {
                    mRiseTime = root2 + hour;
                    mSetTime = root1 + hour;
                } else {
                    mRiseTime = root1 + hour;
                    mSetTime = root2 + hour;
                }
                mRises = true;
                mSets = true;
            }          

            y_minus = y_plus; // prepare for next interval
            hour += 2;         
        } while (!((hour == 25) || (mRises && mSets)));
    }

    public UtcDate getRiseTime() {
        return mRises ? getUtcDate(mRiseTime) : null;
    }
    
    public UtcDate getSetTime() {
        return mSets ? getUtcDate(mSetTime) : null;
    }
    
    public UtcDate getTransitTime() {
        return (mRises && mSets) ? getUtcDate((mRiseTime + mSetTime) / 2) : null;
    }

    // SinAlt: Sine of the altitude of Sun or Moon
    // Input:
    //   Event     Indicates event to find
    //   MJD0      0h at date to investigate (as Modified Julian Date)
    //   Hour      Hour
    //   lambda    Geographic east longitude in [rad]
    //   Cphi      Cosine of geographic latitude
    //   Sphi      Sine of geographic latitude
    // <return>:   Sine of the altitude of Sun or Moon at instant of Event
    private static double SinAlt(double MJD0, int Hour, double lambda, double Cphi, double Sphi) {
        double MJD = MJD0 + Hour / 24.0;

        EclipticPosition ecliptic = Astrometric.sunPosition(MJD);
        EquitorialPosition equitorial = ecliptic.toEquitorialPosition();

        // tau = GMST(MJD) + lambda - RA;
        double tau = UtcDate.GMST(MJD) + lambda - (equitorial.getRightAscension());

        // return (Sphi * sin(Dec) + Cphi * cos(Dec) * cos(tau));
        return Sphi * Math.sin(equitorial.getDeclination())
          + Cphi * Math.cos(equitorial.getDeclination()) * Math.cos(tau);
    }

    private static UtcDate getUtcDate(double time) {
//      double x = Math.abs(Math.floor(time.Mul(60).Add(new Float(5, -1))).Div(60).Add(new Float(1, -5)));
        double x = Math.abs(Math.floor(time * 60 + 0.5) / 60 + 0.00001);
        int H = (int)x;
        x = (x - H) * 60;
        int M = (int)x;
        return new UtcDate(0, 0, 0, H, M, 0);
    }

}
